from __future__ import annotations
from typing import List
import sys, time, pydevd
from PySide2.QtCore import  QObject, QThread, Signal, Slot, QTimer, QEventLoop
from Project.ProjectSettings import ProjectSettings
from Frames.FrameParser import FrameParser
from Frames.Frame import Frame
from Frames.EbdFrame import EbdFrame
from Frames.FrameAddress import FrameAddress
from Utilities.Constants import Constants
from Utilities.Log import Log
from UI.UISignals import UISignals
from Communication.CommandManager import CommandManager
from Communication.SerialPort import SerialPort
import random

class UserApplicationSignals(QObject):
    """This class holds the signals user application
    """
    #: Signal to be called to stop the user application      
    ApplicationStopped = Signal(bool)

class UserApplication(QThread):
    """This class can be used as user application placeholder which could be able to be loaded and executed dynamically
    """

    # #: The serial port to be used by :class:SerialPort
    # __SERIAL_PORT = 'COM9'
    # #: The baud rate for :class:SerialPort
    # __SERIAL_PORT_BAUDRATE = 115200

    #: The serial port to be used by :class:SerialPort for the test results
    __SERIAL_PORT_TEST_REPORT = '/dev/ttyUSB0'
    #: The serial port to be used by :class:SerialPort for reseting the FPGA board
    #__SERIAL_PORT_RESET = 'COM13'
    #: The baud rate for :class:SerialPort
    __SERIAL_PORT_BAUDRATE = 19200
    #: Timeout (in milliseconds) for a test to be completed
    __TEST_COMPLETED_TIMEOUT_MSEC = 1000

    def __init__(self, project : ProjectSettings, frameParser : FrameParser, remoteDeviceSwIdReceived : QObject, parent=None):  
        """Class constructor

        :param project: The project settings
        :type project: ProjectSettings
        :param frameParser: The :class:FrameParser as created by the caller
        :type frameParser: FrameParser
        :param remoteDeviceSwIdReceived: Callbacl function to be called when a SW version and ID are received
        :type remoteDeviceSwIdReceived: QObject
        :param parent: The parent object, defaults to None
        :type parent: [type], optional
        """                  
        super().__init__(parent)
        self._execute = False
        self._project = project
        self._timerTestCompleted = None
        self._testResponseError = False
        self._serialPortTestReport = None
        self._eventLoop = None
        self._frameParser = frameParser
        self._commandManager = None
        self._uiSignals = UISignals()
        self._uiSignals.RemoteDeviceSwIdReceived.connect(remoteDeviceSwIdReceived)

    def Exit(self):
        """Exits the user application
        This method terminates the thread and the closes the :class:CommandManager
        """        
        self.StopClicked()
        try:
            if self._commandManager:
                self._commandManager.Close()              
        except Exception as e:
            Log.PrintException(f'UserApplication.Exit: {str(e)}') 

    @Slot(bool)
    def StopClicked(self, value : bool = True):
        """Stops the thread

        :param value: User provided value (from the parent object), defaults to True
        :type value: bool, optional
        """        
        if self._eventLoop and self._eventLoop.isRunning():
            self._eventLoop.quit()
            self._eventLoop = None
        self._execute = False
        if self.isRunning():
            self.terminate()
        try:
            if self._serialPortTestReport:
                self._serialPortTestReport.Close()          
        except Exception as e:
            Log.PrintException(f'UserApplication.StopClicked: {str(e)}') 

    def __TimerTestCompletedTimeout(self):
        self._testResponseError = True
        if self._eventLoop:
            self._eventLoop.quit()

    @Slot(bytes)
    def TestResultReceived(self, dataBytes : bytes):
        """Callback function for the serial data received from the FPGA when test results are reported

        :param dataBytes: The received bytes
        :type dataBytes: bytes
        """        
        try:
            value = dataBytes.decode('utf-8')
            f=open(self._project.Path+"/results.txt","a+")
            f.write(value+"\n")
            f.close()
            Log.Print(f'UART DATA: {value}')
            self._eventLoop.quit()

        except Exception as e:
            Log.PrintException(f'UserApplication.TestResultReceived: {str(e)}')


    def run(self):
        """This method holds the user implementation executed by this thread
        """
        # pydevd is required for debugging which can be installed with .. code-block:: python pip install pydevd        
        #pydevd.connected = True
        #pydevd.settrace(suspend=False)

        # Open the serial port for the rests results
        self._serialPortTestReport = SerialPort(UserApplication.__SERIAL_PORT_TEST_REPORT, UserApplication.__SERIAL_PORT_BAUDRATE)        
        self._serialPortTestReport.Open(self.TestResultReceived)
        
        try:
            self._commandManager = CommandManager(self._project.FpgaDevice, self._project.IpAddress, self._project.TcpPort)
            self._execute = True

            # Read S/N ans SW version - use allways this mwthod to inform the main application about SW version and remote device S/N
            boardInfo = self._commandManager.ReadId()
            serialNumberHex = '{0:0{1}X}'.format(boardInfo.SerialNumber, 8)
            self._uiSignals.RemoteDeviceSwIdReceived.emit(f'{serialNumberHex} : {boardInfo.Version}')

            
            #Extract fault list
            #Finds all the sensitive bits which are non-masked
            # :return: A list of tuples where each tuple consists of:
            # +-------+------------------------+-----------------------------------------------------------------+
            # | Index | Name                   | Description                                                     |
            # +=======+========================+=================================================================+
            # | 0     | frameIndex             | The index in the provided list where the item resides           |
            # +-------+------------------------+-----------------------------------------------------------------+
            # | 1     | bitIndex               | The bit index (which is sensitive and non-masked) in the frame  |
            # +-------+------------------------+-----------------------------------------------------------------+
            # | 2     | frame address value    | The frame address                                               |
            # +-------+------------------------+-----------------------------------------------------------------+            
            nonMaskedSensitiveBits = EbdFrame.FindNonMaskedSensitiveBits(self._frameParser.EbdFrames, self._frameParser.MskFrames)
            
            TOTAL_BITS_TO_BE_INJECTED=58176 

            frames=[]
            
            f=open(self._project.Path+"/frames.txt","r")
            for line in f:
                tmp=line.replace("\n","")
                frm=int(tmp, 16)
                frames.append(frm)
            f.close()

            dut_non_masked_bits=[]
            for item in nonMaskedSensitiveBits:
               if item[2] in frames:
                   dut_non_masked_bits.append(item)
            
      
            rand_items=random.choices(dut_non_masked_bits,k=TOTAL_BITS_TO_BE_INJECTED)


            test = 0
            #Get the frame address & bit to be injected  
            for item in rand_items:
                Log.Print(f'Test iteration: {str(test + 1)}')

                bitIndex = item[1]
                frameAddress = item[2] 
                
                # Read a frame
                frames = self._commandManager.ReadFrame(frameAddress, 1)
                #Flip the value of the specific bit in the specific frame Address
                frames[0].BitFlip(bitIndex)

                #Inject the fault by writing the specific frame
                status = self._commandManager.WriteFrame(frames)
                Log.Print(str(status))
                    
                self._testResponseError = False
                self._eventLoop = QEventLoop()
                self._timerTestCompleted = QTimer()
                self._timerTestCompleted.setSingleShot(True)
                self._timerTestCompleted.timeout.connect(self.__TimerTestCompletedTimeout)
                self._timerTestCompleted.start(UserApplication.__TEST_COMPLETED_TIMEOUT_MSEC)    

                self._serialPortTestReport.Send("X".encode('utf-8'))
                # Wait for end of command execution or timeout
                self._eventLoop.exec_(QEventLoop.AllEvents)
                # Stop timer if active
                if self._timerTestCompleted and self._timerTestCompleted.isActive():
                    self._timerTestCompleted.stop()
                # Reset the board on timeout
                if self._testResponseError:
                    UserApplication.ResetBoard(self._project.Path)
                    time.sleep(2)
                    continue
                else:
                    # Read a frame
                    frames = self._commandManager.ReadFrame(frameAddress, 1)

                    #correct fault
                    frames[0].BitFlip(bitIndex)
                    status = self._commandManager.WriteFrame(frames)
                    Log.Print(str(status))
                
                test += 1 

        except Exception as e:
            Log.PrintException(f'UserApplication.run: {str(e)}')
        self.Exit()


    @staticmethod
    def ResetBoard(self):
        try:
            if self._serialPortTestReport:
                self._serialPortTestReport.Close()   
            print("RESET THE BOARD!\nPress Enter through command line panel to continue after Reset")
            continue_reset=input("Press Enter when ready to continue")
            self._serialPortTestReport.Open(self.TestResultReceived)
            
        except Exception as e:
            Log.PrintException(f'UserApplication.ResetBoard: {str(e)}')
